import 'dart:async';
import 'dart:typed_data';
import 'dart:ui' as ui;

import 'package:flutter/material.dart';
import 'package:navigine_sdk/navigine_sdk.dart';

class SubLocationWidget extends StatefulWidget {
  final SubLocation subLocation;
  final Function(Venue)? onVenueTapped;
  final VoidCallback? onMapTapped;
  final DeviceInfo? deviceInfo;

  const SubLocationWidget({
    Key? key,
    required this.subLocation,
    this.onMapTapped,
    this.onVenueTapped,
    this.deviceInfo,
  }) : super(key: key);
  @override
  _SubLocationWidgetState createState() => _SubLocationWidgetState();
}

class _SubLocationWidgetState extends State<SubLocationWidget> {
  late SubLocation subLocation;
  Size imageSize = const Size(1, 1);
  Offset? currentPosition;

  @override
  initState() {
    super.initState();
    getImage();
    subLocation = widget.subLocation;
    if (widget.deviceInfo != null) {
      currentPosition = Offset(
        widget.deviceInfo!.x / subLocation.width,
        widget.deviceInfo!.y / subLocation.height,
      );
    }
  }

  Uint8List? image;
  getImage() async {
    image = await widget.subLocation.getImage();
    Completer<ui.Image> completer = Completer<ui.Image>();
    MemoryImage(image!).resolve(const ImageConfiguration()).addListener(
        ImageStreamListener(
            (ImageInfo info, bool _) => completer.complete(info.image)));

    ui.Image i = await completer.future;

    setState(() {
      imageSize = Size(i.width.toDouble(), i.height.toDouble());
    });
  }

  @override
  void didUpdateWidget(SubLocationWidget oldWidget) {
    super.didUpdateWidget(oldWidget);
    setState(() {
      if (widget.subLocation != oldWidget.subLocation) {
        subLocation = widget.subLocation;
        getImage();
      }
      if (widget.deviceInfo != null) {
        currentPosition = Offset(
          widget.deviceInfo!.x / subLocation.width,
          widget.deviceInfo!.y / subLocation.height,
        );
      }
    });
  }

  Venue? shouldInvokeTap(Offset tapped, BuildContext context) {
    var size = MediaQuery.of(context).size;
    var widgetHeight = size.width / imageSize.width * imageSize.height;
    for (var b in subLocation.venues) {
      var x = b.x! / subLocation.width * size.width;
      var y = (1 - b.y! / subLocation.height) * widgetHeight;
      if ((tapped.dx - x).abs() < 10 && (tapped.dy - y).abs() < 10) {
        return b;
      }
    }
    return null;
  }

  double originalScale = 1;
  double flexibleScale = 1;
  final TransformationController _controller = TransformationController();
  @override
  Widget build(BuildContext context) {
    var size = MediaQuery.of(context).size;
    if (image == null) return SizedBox();
    return InteractiveViewer(
      transformationController: _controller,
      minScale: 1,
      maxScale: 50,
      clipBehavior: Clip.none,
      onInteractionUpdate: (s) {
        setState(() {
          flexibleScale = (originalScale * s.scale).clamp(1, 50);
        });
      },
      onInteractionEnd: (s) {
        originalScale = flexibleScale;
      },
      child: Builder(builder: (context) {
        return Container(
          alignment: Alignment.center,
          height: double.infinity,
          child: Stack(
            children: <Widget>[
              SizedBox(
                width: size.width,
                height: size.width / imageSize.width * imageSize.height,
                child: GestureDetector(
                  onTapDown: (detail) {
                    RenderBox? box = context.findRenderObject() as RenderBox?;
                    Offset? local = box?.globalToLocal(detail.globalPosition);
                    if (local != null) {
                      Venue? venue = shouldInvokeTap(local, context);
                      if (venue != null) {
                        widget.onVenueTapped?.call(venue);
                      } else {
                        widget.onMapTapped?.call();
                        _controller.toScene(detail.localPosition);
                      }
                    }
                  },
                  child: Stack(
                    clipBehavior: Clip.antiAlias,
                    children: <Widget>[
                      Image.memory(
                        image!,
                        width: imageSize.width,
                      ),
                      CustomPaint(
                        size: Size(
                          size.width,
                          size.width / imageSize.width * imageSize.height,
                        ),
                        painter: _RoutePainter(
                          imageSize: Size(subLocation.width.toDouble(),
                              subLocation.height.toDouble()),
                          scale: flexibleScale,
                          paths: widget.deviceInfo?.paths
                              .map<List<Offset>>(
                                (e) => e.points
                                    .map(
                                      (e) => Offset(
                                        e.x / subLocation.width,
                                        e.y / subLocation.height,
                                      ),
                                    )
                                    .toList(),
                              )
                              .toList(),
                          position: currentPosition,
                          venues: subLocation.venues
                              .map(
                                (e) => Offset(
                                  e.x! / subLocation.width,
                                  e.y! / subLocation.height,
                                ),
                              )
                              .toList(),
                        ),
                      ),
                    ],
                  ),
                ),
              ),
              if (currentPosition?.dx == 0 && currentPosition?.dy == 0)
                const Center(child: CircularProgressIndicator(strokeWidth: 2)),
            ],
          ),
        );
      }),
    );
  }
}

class _RoutePainter extends CustomPainter {
  // This contains x,y for the users current location
  Offset? position;
  // THis is the size of the image, I am paintiing
  Size imageSize;

  List<List<Offset>>? paths;

  double scale;

  // This is the painting style for beacon dot
  var venuePainter = Paint()..color = Colors.red;

  // This is the painting style for live user dot
  var dotPainter = Paint()..color = Colors.blue;

  // This is the painting style for route line

  List<Offset> venues = [];

  _RoutePainter({
    this.venues = const [],
    this.position,
    this.paths,
    this.scale = 1,
    required this.imageSize,
  });

  Offset getConvertedPoint(Offset point, Size size) {
    return Offset(point.dx * size.width, (1 - point.dy) * size.height);
  }

  @override
  void paint(Canvas canvas, Size size) {
    var routePainter = Paint()
      ..strokeWidth = 2 / scale
      ..style = PaintingStyle.stroke
      ..color = Colors.red;
    int i = 0;
    paths?.forEach((points) {
      Path path = Path();
      // Drawing line using a path, first point is starting the path and else are drawing line to the path
      for (Offset p in points) {
        // Navigine uses dy in opposite direction, thus this fuction does 1- dy and multiplies it by the width and height of the image.
        Offset converted = getConvertedPoint(p, size);
        if (i == 0) {
          path.moveTo(converted.dx, converted.dy);
        } else {
          path.lineTo(converted.dx, converted.dy);
        }
        i++;
      }

      // Drawing the line after all the points, this will rise up the line ie z index is high
      canvas.drawPath(path, routePainter);
    });

    // Drawing circle on the map for each beacons

    for (Offset venue in venues) {
      Offset position = Offset(
        venue.dx * size.width,
        (1 - venue.dy) * size.height,
      );
      canvas.drawCircle(
        position,
        4 / scale,
        venuePainter..color = Colors.pink,
      );
    }

    // Drawing circle on the map for live user location
    if (position != null && position!.dy != 0 && position!.dx != 0) {
      canvas.drawCircle(
        Offset(position!.dx * size.width, (1 - position!.dy) * size.height),
        8 / scale,
        dotPainter,
      );
    }
  }

  @override
  bool shouldRepaint(_RoutePainter oldDelegate) {
    return true;
  }

  List<ui.Offset> decodePolylineToOffset(String polyline) {
    return [];
  }
}

class BeaconDot {
  final String name;
  final double x, y;
  final Color color;

  BeaconDot({
    required this.name,
    required this.x,
    required this.y,
    required this.color,
  });
}
